#include <sta/config.hpp>
#ifdef STA_TACOS_CAN_BUS_ENABLED

#include <sta/tacos/can_bus.hpp>
#include <sta/debug/assert.hpp>
#include <sta/tacos/statemachine.hpp>
#include <sta/tacos.hpp>

extern CAN_HandleTypeDef STA_STM32_CAN_HANDLE;

namespace sta
{
	namespace tacos
	{
		CanBus::CanBus()
			: TacosThread{"Can Bus", STA_TACOS_CAN_BUS_PRIORITY},
			  canBusController_(new STM32CanController(&STA_STM32_CAN_HANDLE)),
			  canBusSysQueue_(STA_RTOS_CAN_BUS_QUEUE_LENGTH),
			  canBusRxQueue_(STA_RTOS_CAN_BUS_QUEUE_LENGTH),
			  canBus_{canBusController_, HAL_GetTick}
		{
		}

		void CanBus::init()
		{
			canBusController_->start();

			if (HAL_CAN_ActivateNotification(&STA_STM32_CAN_HANDLE, CAN_IT_RX_FIFO0_MSG_PENDING) != HAL_OK ||
				HAL_CAN_ActivateNotification(&STA_STM32_CAN_HANDLE, CAN_IT_RX_FIFO1_MSG_PENDING))
			{
				Error_Handler();
			}
		}

		void CanBus::func()
		{
			messageEvent.clear(STA_RTOS_CAN_ANY);
			uint32_t flags = messageEvent.wait(STA_RTOS_CAN_ANY, osWaitForever);

			if (flags != static_cast<uint32_t>(osErrorTimeout))
			{
				STA_ASSERT_MSG((flags & osStatusReserved) == flags, "Unexpected error occurred in wait");

				if (flags & STA_RTOS_CAN_FLAG_SYS_QUEUED)
				{
					// Take messages from queue until empty
					CanSysMsg msg;
					while (CanBus::_instance->getCanBusMsg(&msg, 0))
					{
						canBus_.send(msg);
					}
				}

				if (flags & STA_RTOS_CAN_FLAG_MSG_AVAIL)
				{
					CanSysMsg sysMsg;
					// Iterate through buffer and set back to nullptr after use
					while (canBusRxQueue_.get(&sysMsg, 0))
					{
#ifndef STA_CAN_BUS_FWD_ENABLE
						handleSysMessage(sysMsg.header, sysMsg.payload);
#else
						if (!handleSysMessage(sysMsg.header, sysMsg.payload))
						{

							// Append to the correct thread's queue
							for (std::shared_ptr<TacosThread> thread : Statemachine::instance()->getActiveThreads())
							{
								if (thread->getCanID() == sysMsg.header.sid)
								{
									thread->CAN_queue_.put(sysMsg);
									break;
								}
							}
						}
#endif // STA_CAN_BUS_FWD_ENABLE
					}
				}
			}
		}

		bool CanBus::queueCanBusMsg(const CanSysMsg msg, uint32_t timeout)
		{
			// This technically should check if we are using a system message, but we just pretending that everything is one of those rn
			// STA_ASSERT((msg.header.sid & ~STA_CAN_SID_SYS_BITS) == 0);

			if (canBusSysQueue_.put(msg, timeout))
			{
				// Signal task
				messageEvent.set(STA_RTOS_CAN_FLAG_SYS_QUEUED);
				messageEvent.clear(STA_RTOS_CAN_ANY);
				return true;
			}
			else
			{
				return false;
			}
		}

		void CanBus::canCallback(uint32_t fifo)
		{
			if (messageEvent.get() != STA_RTOS_CAN_FLAG_MSG_AVAIL)
			{
				// get here does not work since FreeRTOS is a buggy mess
				messageEvent.set(STA_RTOS_CAN_FLAG_MSG_AVAIL);

				CanRxHeader rxHeader;						 // CAN Bus Receive Header
				uint8_t canRX[8] = {0, 0, 0, 0, 0, 0, 0, 0}; // CAN Bus Receive Buffer

				bool received_ = canBusController_->receiveFrame(fifo, &rxHeader, canRX);

				if (received_ && rxHeader.id.sid <= 255)
				{
					CanSysMsg sysMsg;
					sysMsg.header.sid = rxHeader.id.sid;
					sysMsg.header.format = 0;

					sysMsg.header.payloadLength = rxHeader.payloadLength;

					for (int i = 0; i < rxHeader.payloadLength; i++)
					{
						sysMsg.payload[i] = canRX[i];
					}

					canBusRxQueue_.put(sysMsg, 0);
				}
			}
		}

		bool CanBus::getCanBusMsg(CanSysMsg *msg, uint32_t timeout)
		{
			return (canBusSysQueue_.get(msg, timeout));
		}

		CanBus *CanBus::_instance = nullptr;

		RtosEvent CanBus::messageEvent;

	} /* namespace tacos */
} /* namespace sta */

namespace sta
{

	void CanBus_RxPendingCallback(uint32_t fifo)
	{
		sta::tacos::CanBus::instance()->canCallback(fifo);
	}

	namespace tacos
	{
		STA_WEAK
		bool handleSysMessage(CanMsgHeader &header, uint8_t *payload)
		{
			// This is a weak function that can be overridden by the user,
			// if they want to handle system messages in a different way, i.e. ignore them

			if (header.sid == STA_TACOS_CAN_BUS_SYS_MSG_ID)
			{
				STA_ASSERT(header.payloadLength == 2);

				// First byte of payload is the origin state, second byte is the destination state. Transition is forced
				tacos::forceState(payload[0], payload[1], 0, true);

				return true;
			}

			return false;
		}
	} // namespace tacos
} // namespace sta

#endif // STA_TACOS_CAN_BUS_ENABLED
